%%
%% Generate a theory file and a proof file for a sts theory.
%%
%% gen_sts_theory File TheoryName
%% uses the trans facts in File to generate a theory called TheoryName.
%% The file TheoryName.thy contains the declarations for the theory and
%% {TheoryName}_proofs.thy contains a collection of proof scripts for the
%% move theorems.
%%

%% Op decls that parallel ergo's decls
?-op(480, fy, not).
?-op(520, xfy, and).
?-op(530, xfy, or).
?-op(540, xfy, =>).
?-op(550, xfy, <=>).


main([File,TName|_]) :-
    atom_concat(TName, '.thy', TFile),
    once(load_the_file(File)),
    check_trans_info,
    open(TFile, write, TS),
    write_header(TS, TName),
    make_const_decls(TS),
    make_trans_axioms(TS),
    make_to_form_rules(TS),
    make_state_axioms(TS),
    write(TS, '\nclose_theory.\n\n'),
    write(TS, '\nsave.\n\n'),
    close(TS).

load_the_file(File) :-
    open(File, read, S),
    repeat,
    read(S, T),
    (   T == end_of_file
        ->
          close(S)
        ;
          assert(T),
          fail
    ).

check_trans_info :-
    trans(L,F,T,V,S1,S2),
    trans(L,F1,T1,V1,S11,S21),
    (   trans(L,F,T,V,S1,S2) = trans(L,F1,T1,V1,S11,S21)
        ->
          fail
        ;
          write_term_list(stderr, [trans(L,F,T,V,S1,S2), ' and ', 
                     trans(L,F1,T1,V1,S11,S21), 'have the same label' , nl]),
         !,fail
    ).
check_trans_info.
    
write_header(TS, TName) :-
    write_term_list(TS, ['%% STS theory : ', TName, nl]),
    write_term_list(TS, ['[sts].', nl]),
    write_term_list(TS, ['delete_theory_and_dependents(', TName, ').', nl]),
    write_term_list(TS, ['add_theory(', TName, ').', nl]),
    write(TS, '\ninclude_theory(stdops, share).\n'),
    write(TS, 'include_theory(iarith, all).\n'),
    write(TS, 'include_theory(localness, all).\n'),
    write(TS, 'include_theory(sts, all).\n'),
    write(TS, '\ndeclare_object_vars(w).\n'),
    write(TS, 'declare_object_vars(y).\n'),
    write(TS, 'declare_object_vars(z).\n\n').

make_const_decls(TS) :-
    setof(A, theory_consts(A), AL),
    make_const_decls(AL, TS).

theory_consts(A) :-
    trans(A, _, _, _, _, _).
theory_consts(A) :-
    trans(_, _, A, _, _, _).
theory_consts(A) :-
    trans(_, _, _, _, state(A,_), _).
theory_consts(A) :-
    trans(_, _, _, _, _, state(A,_)).

a_state(X) :-
    trans(_, _, _, _, state(X,_), _).
a_state(X) :-
    trans(_, _, _, _, _, state(X,_)).

transition(Y) :-
    trans(_, _, Y, _, _, _).

make_const_decls([], _).
make_const_decls([H|T], TS) :-
    write_term_list(TS, ['declare ', H, '.', nl]),
    make_const_decls(T, TS).

make_trans_axioms(TS) :-
    setof(X, a_full_state(X), XL),
    setof(Y, transition(Y), TL),
    make_trans_axioms_aux(TS, XL, TL).

label(Lab, T, S) :-
    trans(Lab, _, T, _, state(S,_), _).

label_set(Labels, T, S) :-
    setof(Lab, label(Lab, T, S), Labels),
    !.
label_set([], _, _).

a_label(Lab) :-
    trans(Lab, _, _, _, _, _).

make_trans_axioms_aux(TS, XL, TL) :-
    member(state(X,N), XL),
    member(T, TL),
    concat_atom([is_trans, X, T], '_', RN),
    label_set(TLL, T, X),
    length(XS,N),
    write_term_list(TS, ['axiom ', RN, ' === ', 
                         wR(is_trans(state(X,XS), T, V)), ' <=> ']),
    write_RHS(TLL, TS, V),
    write_term_list(TS, ['.', nl]),
    fail.
make_trans_axioms_aux(_,_,_).

write_RHS([], TS, _) :-
    write_term_list(TS, [false]).
write_RHS([TL], TS, V) :-
    !,
    write_term_list(TS, ['(', wR(V), ' = ', TL, ')']).
write_RHS([TL|TLL], TS, V) :-
    write_term_list(TS, ['(', wR(V), ' = ', TL, ') or ']),
    write_RHS(TLL, TS, V).

make_to_form_rules(TS) :-
    setof(Lab, a_label(Lab), Labels),
    make_to_form_rules_aux(TS, Labels).

make_to_form_rules_aux(TS, Labels) :-
    member(L, Labels),
    trans(L, F, _, FV, From, TO),
    concat_atom([to, L], '_', RN),
    R = (to(L, From, FV) = TO),
    write_term_list(TS, ['axiom ', RN, ' === ', wR(R), '.', nl]),
    concat_atom([formula, L], '_', FRN),
    FTerm = formula(L, From, FV),
    write_term_list(TS, ['axiom ', FRN, ' === ', 
                         wR(FTerm), ' = (', wR(F), ')', '.', nl]),
    fail.
make_to_form_rules_aux(_, _).
    
    
a_full_state(state(S,N)) :-
    trans(_, _, _, _, state(S,L), _),
    length(L,N).
a_full_state(state(S,N)) :-
    trans(_, _, _, _, _, state(S,L)),
    length(L,N).

make_state_axioms(TS) :-
    setof(Term, a_full_state(Term), Terms),
    make_state_axioms_aux(TS, Terms).

make_state_axioms_aux(TS, Terms) :-
    member(state(S, N), Terms),
    length(L, N),
    concat_atom([is_state, S], '_', RN),
    write_term_list(TS, ['axiom ', RN, ' === ', wR(is_state(state(S, L))), '.', nl]),
    fail.
make_state_axioms_aux(_, _).

